
//
// Copyright (C) 2016 by Tobias Glasmachers.
// This file is part of the supplementary material for the paper:
//
// Tobias Glasmachers. Finite Sum Acceleration vs. Adaptive Learning
// Rates for the Training of Kernel Machines on a Budget. NIPS workshop
// on Optimization for Machine Learning, 2016.
//
// The program depend on version 3.1 (or later) of the Shark machine
// learning library (shark-ml.org).
//
//
// LICENSE
// -------
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this file, to copy and distribute it freely, provided that
// the header of comments including this license text is contained in
// all copies made.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//


// shark includes
#include <shark/LinAlg/Base.h>
#include <shark/Data/Dataset.h>
#include <shark/Data/SparseData.h>
#include <shark/ObjectiveFunctions/Loss/AbstractLoss.h>
#include <shark/ObjectiveFunctions/Loss/SquaredHingeLoss.h>
#include <shark/ObjectiveFunctions/Loss/CrossEntropy.h>
#include <shark/ObjectiveFunctions/Loss/SquaredLoss.h>
#include <shark/ObjectiveFunctions/Loss/ZeroOneLoss.h>
#include <shark/Models/Kernels/GaussianRbfKernel.h>
#include <shark/Models/Normalizer.h>
#include <shark/Algorithms/Trainers/NormalizeComponentsUnitVariance.h>

// standard includes
#include <vector>
#include <set>
#include <map>
#include <string>
#include <iostream>
#include <stdexcept>
#include <limits>
#include <algorithm>
#include <cmath>


using namespace shark;
using namespace std;


// uncomment exactly one of the two:

//#define REGRESSION
//typedef RealVector LabelType;
//typedef SquaredLoss<RealVector, RealVector> CostFunction;

#define CLASSIFICATION
typedef unsigned int LabelType;
typedef ZeroOneLoss<unsigned int, RealVector> CostFunction;


// types
typedef CompressedRealVector InputType;
typedef LabeledData<InputType, LabelType> Dataset;
typedef Data<InputType> BasisType;
typedef typename BasisType::const_element_reference InputReference;
typedef typename BasisType::const_batch_reference InputBatch;
typedef AbstractLoss<LabelType, RealVector> LossFunction;
typedef Batch<LabelType>::type LabelBatch;


// lots of "bad" global variables
size_t B = 1000;                       // budget, number of support vectors
GaussianRbfKernel<InputType> kernel;   // can be any class derived from shark::AbstractKernelFunction
RealVector alpha;                      // vector of coefficients
BasisType basis;                       // "support" vectors
Dataset traindata;
Dataset testdata;
LossFunction* loss;                    // convex training loss function
CostFunction cost;                     // evaluation loss function
double lambda;                         // regularization parameter
size_t n;                              // number of training points
size_t n2;                             // number of test points
size_t m;                              // number of basis functions
size_t epochs;
RealMatrix basisK;                     // kernel Gram matrix of the basis
RealVector basisK2;                    // squared norms of row vectors
RealMatrix dataK;                      // kernel Gram matrix data x basis
RealVector dataK2;                     // squared norms of row vectors, negative if not computed yet
RealMatrix testK;                      // kernel Gram matrix test data x basis
LabelBatch trainlabels;                // random access to labels
LabelBatch testlabels;                 // random access to labels


// Compute the primal objective value. This is needed for evaluation only.
double primal()
{
	double squarednorm = blas::inner_prod(alpha, blas::prod(basisK, alpha));
	RealMatrix predictions(n, 1);
	blas::column(predictions, 0) = blas::prod(dataK, alpha);
	double empiricalrisk = (*loss)(trainlabels, predictions) / n;
	return 0.5 * lambda * squarednorm + empiricalrisk;
}


////////////////////////////////////////////////////////////
// Five different optimization procedures for training.
//

// Gradient Descent (GD) training
bool trainGD(double eta)
{
	cout << "GD training" << endl;

	alpha.clear();

	double baseline = primal();
	for (size_t epoch=0; epoch<epochs; epoch++)
	{
		RealMatrix predictions(n, 1);
		blas::column(predictions, 0) = blas::prod(dataK, alpha);
		RealMatrix lossgradient;
		loss->evalDerivative(trainlabels, predictions, lossgradient);
		RealVector gradient = blas::prod(blas::trans(dataK), blas::column(lossgradient, 0));
		gradient /= n;
		RealVector reggrad = lambda * blas::prod(basisK, alpha);
		gradient += reggrad;

		alpha -= eta * gradient;

		// record statistics after each epoch
		cout << "[" << epoch << flush;
		double obj = primal();
		cout << ", " << obj << flush;
		RealMatrix predictions2(n2, 1);
		blas::column(predictions2, 0) = blas::prod(testK, alpha);
		double testerror = cost(testlabels, predictions2) / n2;

		cout << ", " << testerror << "]," << endl;

		if (! isfinite(obj) || obj > baseline) return false;   // safeguard against wasting time on diverging run
	}
	return true;
}

// Stochastic Gradient Descent (SGD) training
bool trainSGD(double eta)
{
	cout << "SGD training" << endl;

	alpha.clear();
	RealVector prediction(1);
	RealVector lossgradient(1);

	double baseline = primal();
	for (size_t epoch=0; epoch<epochs; epoch++)
	{
		for (size_t iter=0; iter<n+m; iter++)
		{
			const double eta_t = (n+m) * eta / ((n+m)*(epoch+1) + iter);
			size_t b = Rng::discrete(0, n + m - 1);

			if (b < n)
			{
				// make a step on a loss term
				auto k = blas::row(dataK, b);

				// compute prediction
				prediction(0) = blas::inner_prod(k, alpha);

				// compute stochastic gradient of the loss
				loss->evalDerivative(get(trainlabels, b), prediction, lossgradient);
				double derivative = lossgradient(0) / n;

				// update alpha
				alpha -= (eta_t * (n + m) * derivative) * k;
			}
			else
			{
				// make a step on a regularization term
				b -= n;
				auto k = blas::row(basisK, b);

				// compute stochastic gradient of the regularization term
				double derivative = lambda * alpha(b);

				// update alpha
				alpha -= (eta_t * (n + m) * derivative) * k;
			}
		}

		// record statistics after each epoch
		cout << "[" << epoch << flush;
		double obj = primal();
		cout << ", " << obj << flush;
		RealMatrix predictions2(n2, 1);
		blas::column(predictions2, 0) = blas::prod(testK, alpha);
		double testerror = cost(testlabels, predictions2) / n2;
		cout << ", " << testerror << "]," << endl;

		if (! isfinite(obj) || obj > baseline) return false;   // safeguard against wasting time on diverging run
	}
	return true;
}

// Stochastic Variance Reduced Gradient (SVRG) training
bool trainSVRG(double eta)
{
	cout << "SVRG training" << endl;

	alpha.clear();
	RealVector prediction(1);
	RealVector lossgradient(1);

	size_t epoch = 0;
	double baseline = primal();

	RealVector proxy = alpha;   // \tilde w
	RealVector mu(m, 0.0);
	{
		RealMatrix predictions(n, 1);
		blas::column(predictions, 0) = blas::prod(dataK, proxy);
		RealMatrix lossgradient;
		loss->evalDerivative(trainlabels, predictions, lossgradient);
		mu = blas::prod(blas::trans(dataK), blas::column(lossgradient, 0)) / n;
		mu += lambda * blas::prod(basisK, proxy);

		// record statistics after each epoch
		cout << "[" << epoch << flush;
		epoch++;

		double obj = primal();
		cout << ", " << obj << flush;
		RealMatrix predictions2(n2, 1);
		blas::column(predictions2, 0) = blas::prod(testK, alpha);
		double testerror = cost(testlabels, predictions2) / n2;
		cout << ", " << testerror << "]," << endl;
	}
	size_t fullEpoch = epoch + 2;

	for ( ; epoch<epochs; epoch++)
	{
		if (epoch == fullEpoch)
		{
			// update the proxy point and its gradient
			proxy = alpha;
			RealMatrix predictions(n, 1);
			blas::column(predictions, 0) = blas::prod(dataK, proxy);
			RealMatrix lossgradient;
			loss->evalDerivative(trainlabels, predictions, lossgradient);
			mu = blas::prod(blas::trans(dataK), blas::column(lossgradient, 0)) / n;
			mu += lambda * blas::prod(basisK, proxy);

			// record statistics after each epoch
			epoch++;
			fullEpoch = epoch + 2;
		}
		if (epoch >= epochs) break;

		for (size_t iter=0; iter<n+m; iter++)
		{
			size_t b = Rng::discrete(0, n + m - 1);

			if (b < n)
			{
				// make a step on a loss term
				auto k = blas::row(dataK, b);

				// compute prediction
				prediction(0) = blas::inner_prod(k, alpha);

				// compute stochastic gradient of the loss
				loss->evalDerivative(get(trainlabels, b), prediction, lossgradient);
				double derivative = lossgradient(0) / n;

				// compute prediction on proxy
				prediction(0) = blas::inner_prod(k, proxy);

				// compute stochastic gradient of the loss
				loss->evalDerivative(get(trainlabels, b), prediction, lossgradient);
				derivative -= lossgradient(0) / n;

				// update alpha
				alpha -= eta * ((n + m) * derivative * k + mu);
			}
			else
			{
				// make a step on a regularization term
				b -= n;
				auto k = blas::row(basisK, b);

				// compute stochastic gradient of the regularization term
				double derivative = lambda * alpha(b);

				// compute stochastic gradient on the proxy
				derivative -= lambda * proxy(b);

				// update alpha
				alpha -= eta * ((n + m) * derivative * k + mu);
			}
		}

		// record statistics after each epoch
		cout << "[" << epoch << flush;
		double obj = primal();
		cout << ", " << obj << flush;
		RealMatrix predictions2(n2, 1);
		blas::column(predictions2, 0) = blas::prod(testK, alpha);
		double testerror = cost(testlabels, predictions2) / n2;
		cout << ", " << testerror << "]," << endl;

		if (! isfinite(obj) || obj > baseline) return false;   // safeguard against wasting time on diverging run
	}

	return true;
}

// Stochastic Average Gradient (SAG) training
// with online estimation of the Lipschitz constant
bool trainSAG(double eta)
{
	cout << "SAG training" << endl;

	alpha.clear();

	double L = 1.0 / eta;
	vector<double> Li(n + m, L);
	map<double, std::size_t> orderedLi;
	orderedLi[L] = n + m;
	vector<double> memory(n + m, 0.0);
	RealVector g(m);
	RealVector prediction(1);
	RealVector lossgradient(1);

	double baseline = primal();
	for (size_t epoch=0; epoch<epochs; epoch++)
	{
		for (size_t iter=0; iter<n+m; iter++)
		{
			size_t b = Rng::discrete(0, n + m - 1);

			if (b < n)
			{
				// upload the gradient of a loss term
				auto k = blas::row(dataK, b);
				double k2 = dataK2(b);

				// compute stochastic gradient of the loss
				double scp = blas::inner_prod(k, alpha);
				prediction(0) = scp;
				loss->evalDerivative(get(trainlabels, b), prediction, lossgradient);
				double derivative = lossgradient(0) / n;

				// update Lipschitz "constant" (actually an online estimate)
				if (derivative * derivative * k2 > 1e-8)
				{
					auto it = orderedLi.find(Li[b]);
					assert(it != orderedLi.end());
					if (it->second == 1) orderedLi.erase(it);
					else it->second--;

					Li[b] *= 0.5;
					prediction(0) = scp - derivative * k2 / Li[b];
					loss->evalDerivative(get(trainlabels, b), prediction, lossgradient);
					double derivative2 = lossgradient(0) / n;
					while (derivative2 * derivative < 0.0)
					{
						Li[b] *= 2.0;
						prediction(0) = scp - derivative * k2 / Li[b];
						loss->evalDerivative(get(trainlabels, b), prediction, lossgradient);
						derivative2 = lossgradient(0) / n;
					}
					it = orderedLi.find(Li[b]);
					if (it == orderedLi.end()) orderedLi[Li[b]] = 1;
					else it->second++;

					it = orderedLi.end();
					--it;
					L = it->first;
				}

				// update sum of gradients
				g += (derivative - memory[b]) * k;
				memory[b] = derivative;
			}
			else
			{
				// update the gradient of a regularization term
				auto k = blas::row(basisK, b - n);
				double k2 = basisK2(b - n);

				// compute stochastic gradient of the regularization term
				double derivative = lambda * alpha(b - n);

				// update Lipschitz "constant" (actually an online estimate)
				if (derivative * derivative * k2 > 1e-8)
				{
					auto it = orderedLi.find(Li[b]);
					assert(it != orderedLi.end());
					if (it->second == 1) orderedLi.erase(it);
					else it->second--;

					Li[b] *= 0.5;
					double derivative2 = lambda * (alpha(b - n) - (n + m) * derivative / Li[b]);
					while (derivative2 * derivative < 0.0)
					{
						Li[b] *= 2.0;
						derivative2 = lambda * (alpha(b - n) - (n + m) * derivative / Li[b]);
					}
					it = orderedLi.find(Li[b]);
					if (it == orderedLi.end()) orderedLi[Li[b]] = 1;
					else it->second++;

					it = orderedLi.end();
					--it;
					L = it->first;
				}

				// update sum of gradients
				g += (derivative - memory[b]) * k;
				memory[b] = derivative;
			}

			// update alpha
			alpha -= (0.0625 / L) * g;
		}

		// record statistics after each epoch
		cout << "[" << epoch << flush;
		double obj = primal();
		cout << ", " << obj << flush;
		RealMatrix predictions2(n2, 1);
		blas::column(predictions2, 0) = blas::prod(testK, alpha);
		double testerror = cost(testlabels, predictions2) / n2;
		cout << ", " << testerror << "]," << endl;

		if (! isfinite(obj) || obj > baseline) return false;   // safeguard against wasting time on diverging run
	}

	return true;
}


// Tabular distribution based on unnormalized non-negative preference
// values. Construction takes O(n) time, sampling and updating work in
// O(log(n)) time.
class PreferenceBasedDistribution
{
public:
	PreferenceBasedDistribution(size_t n, double value)
	: cumulation(n, 0.0)
	, preference(n, value)
	, sum(n * value)
	{
		assert(n > 0);
		assert(sum > 0.0);
		for (log2n = 1; 2 * log2n < n; log2n *= 2);
		init(0, log2n);
	}

	size_t size() const
	{ return preference.size(); }

	double operator [] (size_t index) const
	{ return preference[index]; }

	double probability(size_t index) const
	{ return (preference[index] / sum); }

	size_t sample() const
	{
		double r = Rng::uni(0.0, sum);
		const size_t n = size();
		size_t pos = 0;
		for (size_t step = log2n; step > 0; step /= 2)
		{
			if (pos + step < n && r >= cumulation[pos + step])
			{
				pos += step;
				r -= cumulation[pos];
			}
		}
		return pos;
	}

	void set(size_t index, double value)
	{
		double diff = value - preference[index];
		sum += diff;
		preference[index] = value;
		assert(sum > 0.0);

		const size_t n = size();
		size_t pos = 0;
		for (size_t step = log2n; step > 0; step /= 2)
		{
			if (pos + step < n)
			{
				if (pos + step >= index) pos += step;
				else cumulation[pos + step] += diff;
			}
		}
	}

private:
	double init(size_t pos, size_t step)
	{
		if (pos >= size()) return 0.0;
		if (step == 0) return preference[pos];
		double l = init(pos, step / 2);
		double r = init(pos + step, step / 2);
		if (pos + step < size()) cumulation[pos + step] = l;
		return (l + r);
	}

	vector<double> cumulation;         // sums of subsets of preference values
	vector<double> preference;         // preference values, unnormalized probabilities
	double sum;                        // sum over all preference values, normalization constant
	size_t log2n;                      // largest step size for binary search
};


// Stochastic Gradient Descent (SGD) training with the ADAM method
bool trainADAM(double eta)
{
	cout << "SGD-ADAM training" << endl;

	alpha.clear();
	RealVector prediction(1);
	RealVector lossgradient(1);

	constexpr double beta1 = 0.9;
	constexpr double beta2 = 0.999;
	constexpr double epsilon = 1e-8;
	double beta1p = 1.0;
	double beta2p = 1.0;
	RealVector mean(alpha.size(), 0.0);
	RealVector variance(alpha.size(), 0.0);

	double baseline = primal();
	for (size_t epoch=0; epoch<epochs; epoch++)
	{
		for (size_t iter=0; iter<n+m; iter++)
		{
			beta1p *= beta1;
			beta2p *= beta2;
			const double eta_t = eta * std::sqrt(1.0 - beta2p) / (1.0 - beta1p);

			size_t b = Rng::discrete(0, n + m - 1);

			double derivative;
			auto k = (b < n) ? blas::row(dataK, b) : blas::row(basisK, b - n);
			if (b < n)
			{
				// make a step on a loss term

				// compute prediction
				prediction(0) = blas::inner_prod(k, alpha);

				// compute stochastic gradient of the loss
				loss->evalDerivative(get(trainlabels, b), prediction, lossgradient);
				derivative = lossgradient(0) / n;
			}
			else
			{
				// make a step on a regularization term

				// compute stochastic gradient of the regularization term
				derivative = lambda * alpha(b - n);
			}

			// update alpha
			mean *= beta1;
			variance *= beta2;
			for (size_t i=0; i<alpha.size(); i++)
			{
				double d = (n + m) * derivative * k(i);
				mean(i) += (1.0 - beta1) * d;
				variance(i) += (1.0 - beta2) * d * d;
				alpha(i) -= eta_t * mean(i) / (std::sqrt(variance(i)) + epsilon);
			}
		}

		// record statistics after each epoch
		cout << "[" << epoch << flush;
		double obj = primal();
		cout << ", " << obj << flush;
		RealMatrix predictions2(n2, 1);
		blas::column(predictions2, 0) = blas::prod(testK, alpha);
		double testerror = cost(testlabels, predictions2) / n2;
		cout << ", " << testerror << "]," << endl;

		if (! isfinite(obj) || obj > baseline) return false;   // safeguard against wasting time on diverging run
	}
	return true;
}


////////////////////////////////////////////////////////////
// Two methods for constructing the basis of size B
// from a training data set.
//

void constructBasisRandom()
{
	basis = traindata.inputs();
	{
		size_t sz = basis.numberOfElements();
		if (sz > B)
		{
			// Truncate the basis. This amounts to random
			// sub-sampling since the points were shuffled.
			basis.makeIndependent();
			splitAtElement(basis, B);
		}
	}
}

struct TreeNode
{
	TreeNode()
	: left(nullptr)
	, right(nullptr)
	{ }

	~TreeNode()
	{
		delete left;
		delete right;
	}

	TreeNode* findLeaf(InputReference x)
	{
		if (left)
		{
			double l = blas::norm_2(x - pt(leftpoint));
			double r = blas::norm_2(x - pt(rightpoint));
			return ((l < r) ? left : right)->findLeaf(x);
		}
		else return this;
	}

	const TreeNode* findLeaf(InputReference x) const
	{
		if (left)
		{
			double l = blas::norm_2(x - pt(leftpoint));
			double r = blas::norm_2(x - pt(rightpoint));
			return ((l < r) ? left : right)->findLeaf(x);
		}
		else return this;
	}

	static InputReference pt(pair<size_t, size_t> const& indices)
	{ return shark::get(traindata.batch(indices.first), indices.second).input; }

	double computeSpread()
	{
		spread = 0.0;
		if (content.size() <= 1) return spread;
		for (size_t t=0; t<59; t++)
		{
			size_t i = Rng::discrete(0, content.size() - 1);
			size_t j = Rng::discrete(0, content.size() - 2);
			if (j >= i) j++;
			double dist = blas::norm_2(pt(content[i]) - pt(content[j]));
			if (dist >= spread)
			{
				spread = dist;
				leftpoint = content[i];
				rightpoint = content[j];
			}
		}
		return spread;
	}

	void split()
	{
		assert(! left);
		assert(! right);
		left = new TreeNode();
		right = new TreeNode();
		for (size_t i=0; i<content.size(); i++)
		{
			double dl = blas::norm_2(pt(content[i]) - pt(leftpoint));
			double dr = blas::norm_2(pt(content[i]) - pt(rightpoint));
			if (dl < dr) left->content.push_back(content[i]);
			else right->content.push_back(content[i]);
		}
		left->computeSpread();
		right->computeSpread();
	}

	void collectPoints(vector<InputType>& base) const
	{
		if (left)
		{
			if (left->left) left->collectPoints(base);
			else base.push_back(pt(leftpoint));

			if (right->left) right->collectPoints(base);
			else base.push_back(pt(rightpoint));
		}
	}

	TreeNode* left;
	TreeNode* right;
	pair<size_t, size_t> leftpoint;
	pair<size_t, size_t> rightpoint;
	double spread;
	vector< pair<size_t, size_t> > content;
};

void constructBasisBSP()
{
	if (traindata.numberOfElements() <= B)
	{
		basis = traindata.inputs();
		return;
	}

	// construct a hierarchical space partition
	TreeNode root;
	multimap<double, TreeNode*> leaves;
	for (size_t i=0; i<traindata.numberOfBatches(); i++)
	{
		InputBatch bi = traindata.batch(i).input;
		size_t sz = boost::size(bi);
		for (size_t j=0; j<sz; j++) root.content.push_back(make_pair(i, j));
	}
	root.computeSpread();
	leaves.insert(make_pair(root.spread, &root));
	while (leaves.size() < B)
	{
		// split the "most widespread" leaf
		auto it = leaves.end(); --it;
		TreeNode* p = it->second;
		p->split();
		leaves.erase(it);
		leaves.insert(make_pair(p->left->spread, p->left));
		leaves.insert(make_pair(p->right->spread, p->right));
	}

	// add all points corresponding to leaf nodes to the basis
	vector<InputType> base;
	root.collectPoints(base);
	assert(base.size() == B);

	// construct the basis data object
	basis = createDataFromRange(base);
}


////////////////////////////////////////////////////////////
// minimalistic command line help
//

void help(string const& programname)
{
	cout << endl
#ifdef CLASSIFICATION
		<< "compiled for CLASSIFICATION" << endl
#endif
#ifdef REGRESSION
		<< "compiled for REGRESSION" << endl
#endif
		<< "usage: " << programname << " <method> <loss> <dataset> <lambda> <gamma> <epochs> <seed> <options>" << endl
		<< "with" << endl
		<< "    method          can be: GD, SGD, SVRG, SAG, or ADAM" << endl
#ifdef CLASSIFICATION
		<< "    loss            can be: SquaredHinge or CrossEntropy" << endl
#endif
#ifdef REGRESSION
		<< "    loss            can be: SquaredLoss" << endl
#endif
		<< "    dataset         can be: singlefile <datafile>" << endl
		<< "                        or: <trainfile> <testfile>" << endl
		<< "    lambda          non-negative number" << endl
		<< "    gamma           can be: estimate or a positive factor" << endl
		<< "    epochs          positive integer" << endl
		<< "    seed            positive integer" << endl
		<< "options:" << endl
		<< "    --budget <value>         default: 1000" << endl
		<< "    --basistype <value>      can be: random, BSP; default: BSP" << endl
		<< "    --normalize              standardize features to unit variance" << endl
		<< endl;

	exit(0);
}


int main(int argc, char** argv)
{
	cout << setprecision(12);

	// parse command line
	if (argc < 9) help(argv[0]);
	std::string methodname = argv[1];
	std::string lossname = argv[2];
	std::string trainfilename = argv[3];
	std::string testfilename = argv[4];
	lambda = strtod(argv[5], NULL);
	double gamma = strtod(argv[6], NULL);
	epochs = strtol(argv[7], NULL, 10);
	int seed = strtol(argv[8], NULL, 10);
	if (lambda < 0.0)
	{
		cout << "regularization parameter (lambda) must be non-negative!" << endl << endl;
		help(argv[0]);
	}
	if (gamma <= 0.0)
	{
		cout << "kernel parameter (gamma) must be positive!" << endl << endl;
		help(argv[0]);
	}
	string basistype = "BSP";
	bool normalize = false;
	for (int a = 9; a < argc; a++)
	{
		string arg = argv[a];
		if (arg == "--budget")
		{
			a++;
			if (a >= argc) help(argv[0]);
			B = strtol(argv[a], NULL, 10);
		}
		else if (arg == "--basistype")
		{
			a++;
			if (a >= argc) help(argv[0]);
			basistype = argv[a];
		}
		else if (arg == "--normalize")
		{
			normalize = true;
		}
		else help(argv[0]);
	}

	Rng::seed(seed);

	// prepare the loss function
#ifdef CLASSIFICATION
	if (lossname == "SquaredHinge") loss = new SquaredHingeLoss();
	else if (lossname == "CrossEntropy") loss = new CrossEntropy();
#endif
#ifdef REGRESSION
	if (lossname == "SquaredLoss") loss = new SquaredLoss<RealVector, RealVector>();
#endif
	else
	{
		cout << "unknown loss type: " << lossname << endl << endl;
		help(argv[0]);
	}

	// load, shuffle, and split data
	cout << "loading data ..." << flush;
	if (trainfilename == "singlefile")
	{
		importSparseData(traindata, testfilename);
		traindata.shuffle();
		size_t num = traindata.numberOfElements();
		if (num > 20000) num -= 10000;
		else num /= 2;
		testdata = splitAtElement(traindata, num);
	}
	else
	{
		importSparseData(traindata, trainfilename);
		importSparseData(testdata, testfilename, inputDimension(traindata));
		traindata.shuffle();
	}
	cout << " done." << endl;

	// normalize data
	if (normalize)
	{
		cout << "normalizing data ..." << flush;
		bool removeMean = false;
		Normalizer<CompressedRealVector> normalizer;
		NormalizeComponentsUnitVariance<CompressedRealVector> normalizingTrainer(removeMean);
		normalizingTrainer.train(normalizer, traindata.inputs());
		traindata = transformInputs(traindata, normalizer);
		testdata = transformInputs(testdata, normalizer);
		cout << " done." << endl;
	}

	// effective budget size cannot exceed the size of the training set
	B = std::min<size_t>(traindata.numberOfElements(), B);

	// prepare kernel
	kernel.setGamma(gamma);

	// construct the basis
	cout << "constructing basis ..." << flush;
	if (basistype == "random") constructBasisRandom();
	else if (basistype == "BSP") constructBasisBSP();
	else throw runtime_error(" unknown basis construction method");
	cout << " done." << endl;

	// dimensions
	n = traindata.numberOfElements();
	n2 = testdata.numberOfElements();
	size_t p = inputDimension(traindata);
	m = basis.numberOfElements();
	cout << "training points: " << n << endl;
	cout << "    test points: " << n2 << endl;
	cout << "       features: " << p << endl;
	cout << "     basis size: " << m << endl;

	// extract vectors of labels
	trainlabels = Batch<LabelType>::createBatch(get(traindata.batch(0).label, 0), n);
	for (size_t i=0, z=0; i<traindata.numberOfBatches(); i++)
	{
		auto batch = traindata.batch(i);
		for (size_t j=0; j<boost::size(batch); j++, z++)
		{
			get(trainlabels, z) = shark::get(batch, j).label;
		}
	}
	testlabels = Batch<LabelType>::createBatch(get(testdata.batch(0).label, 0), n2);
	for (size_t i=0, z=0; i<testdata.numberOfBatches(); i++)
	{
		auto batch = testdata.batch(i);
		for (size_t j=0; j<boost::size(batch); j++, z++)
		{
			get(testlabels, z) = shark::get(batch, j).label;
		}
	}
#ifdef REGRESSION
	// subtract mean of training data from labels
	// (the mean effectively acts as a bias term)
	double sum = 0.0;
	for (size_t i=0; i<trainlabels.size1(); i++) sum += trainlabels(i, 0);
	double mean = sum / trainlabels.size1();
	for (size_t i=0; i<trainlabels.size1(); i++) trainlabels(i, 0) -= mean;
	for (size_t i=0; i<testlabels.size1(); i++) testlabels(i, 0) -= mean;
#endif

	// pre-compute kernel Gram matrices
	// (this is no a necessity; it speeds up experiments significantly)
	cout << "precomputing kernel matrices ..." << flush;
	basisK = RealMatrix(m, m);
	for (size_t i=0, ii=0; i<basis.numberOfBatches(); i++)
	{
		InputBatch bi = basis.batch(i);
		size_t di = boost::size(bi);
		for (size_t j=0, jj=0; j<basis.numberOfBatches(); j++)
		{
			InputBatch bj = basis.batch(j);
			size_t dj = boost::size(bj);
			blas::subrange(basisK, ii, ii+di, jj, jj+dj) = kernel(bi, bj);
			jj += dj;
		}
		ii += di;
	}
	basisK2 = RealVector(m);
	for (size_t i=0; i<m; i++) basisK2(i) = blas::inner_prod(blas::row(basisK, i), blas::row(basisK, i));

	dataK = RealMatrix(n, m);
	for (size_t i=0, ii=0; i<traindata.numberOfBatches(); i++)
	{
		InputBatch bi = traindata.batch(i).input;
		size_t di = boost::size(bi);
		for (size_t j=0, jj=0; j<basis.numberOfBatches(); j++)
		{
			InputBatch bj = basis.batch(j);
			size_t dj = boost::size(bj);
			blas::subrange(dataK, ii, ii+di, jj, jj+dj) = kernel(bi, bj);
			jj += dj;
		}
		ii += di;
	}
	dataK2 = RealVector(n);
	for (size_t i=0; i<n; i++) dataK2(i) = blas::inner_prod(blas::row(dataK, i), blas::row(dataK, i));

	testK = RealMatrix(n2, m);
	for (size_t i=0, ii=0; i<testdata.numberOfBatches(); i++)
	{
		InputBatch bi = testdata.batch(i).input;
		size_t di = boost::size(bi);
		for (size_t j=0, jj=0; j<basis.numberOfBatches(); j++)
		{
			InputBatch bj = basis.batch(j);
			size_t dj = boost::size(bj);
			blas::subrange(testK, ii, ii+di, jj, jj+dj) = kernel(bi, bj);
			jj += dj;
		}
		ii += di;
	}
	cout << " done." << endl;

	// train the model
	alpha = RealVector(m);
	for (double eta = 1.0; eta > 1e-10; eta /= 2.0)
	{
		cout << "************    trying learning rate " << eta << endl;
		if (methodname == "GD")
		{
			if (trainGD(eta)) break;
		}
		else if (methodname == "SGD")
		{
			if (trainSGD(eta)) break;
		}
		else if (methodname == "SVRG")
		{
			if (trainSVRG(eta)) break;
		}
		else if (methodname == "SAG")
		{
			if (trainSAG(eta)) break;
		}
		else if (methodname == "ADAM")
		{
			if (trainADAM(eta)) break;
		}
		else
		{
			cout << "unknown training method type: " << methodname << endl << endl;
			help(argv[0]);
		}
	}

	// compute final errors
	RealMatrix predictions(n, 1);
	blas::column(predictions, 0) = blas::prod(dataK, alpha);
	double trainerror = cost(trainlabels, predictions) / n;
	RealMatrix predictions2(n2, 1);
	blas::column(predictions2, 0) = blas::prod(testK, alpha);
	double testerror = cost(testlabels, predictions2) / n2;
	cout << "training error: " << trainerror << endl;
	cout << "    test error: " << testerror << endl;
}
